from tinyprog import TinyProg, get_ports
import tinyprog
from six.moves.urllib.request import urlopen
from six.moves import input
import sys
import argparse
import json
from packaging import version
import time

# adapted from http://code.activestate.com/recipes/577058/
def query_user(question, default="no"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """

    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


def strict_query_user(question):
    valid = {"yes": True}

    prompt = " [yes/NO] > "

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        return valid.get(choice, False)


def get_port_by_uuid(device, uuid):
    ports = get_ports(device) + get_ports("1209:2100")
    for port in ports:
        try:
            with port:
                p = TinyProg(port)
                if p.meta.uuid().startswith(uuid):
                    return port

        except:
            pass
    return None


def check_for_new_bootloader():
    #ports = [p[0] for p in comports() if "1d50:6130" in p[2].lower()]

    # download bootloader update information
    #bootloader_update_info = urlopen(m[u"bootmeta"][u"update"] + "/bootloader.json").read()
    return []


def check_for_wrong_tinyfpga_bx_vidpid():
    """
    Some of the TinyFPGA BX boards have the wrong USB VID:PID.  This function
    looks for such boards.
    """
    boards_needing_update = []

    ports = get_ports("1209:2100")
    for port in ports:
        with port:
            p = TinyProg(port)
            m = p.meta.root
            if m[u"boardmeta"][u"name"] == u"TinyFPGA BX":
                boards_needing_update.append(port)

    return boards_needing_update


def check_if_overwrite_bootloader(addr, length, userdata_range):
    ustart = userdata_range[0]
    uend = userdata_range[1]

    if addr < ustart or addr + length >= uend:
        print("")
        print("    !!! WARNING !!!")
        print("")
        print("    The address given may overwrite the USB bootloader. Without the")
        print("    USB bootloader the board will no longer be able to be programmed")
        print("    over the USB interface. Without the USB bootloader, the board can")
        print("    only be programmed via the SPI flash interface on the bottom of")
        print("    the board")
        print("")
        retval = strict_query_user("    Are you sure you want to continue? Type in 'yes' to overwrite bootloader.")
        print("")
        return retval

    return True


def perform_bootloader_update(port):
    uuid = None

    with port:
        p = TinyProg(port)
        m = p.meta.root

        uuid = m[u"boardmeta"][u"uuid"]

        # download bootloader update information
        bootloader_update_info = json.loads(urlopen(m[u"bootmeta"][u"update"] + "/bootloader.json").read())

        # ask permission to update the board
        print("")
        print("    The following update:")
        print("")
        print("        New Version: %s" % bootloader_update_info[u"version"])
        print("        Notes: %s" % bootloader_update_info[u"notes"])
        print("")
        print("    is available for this board:")
        print_board(port, m)
        print("")
        
        if query_user("    Would you like to perform the update?"):
            ####################
            # STAGE ONE 
            ####################
            print("    Fetching stage one...")
            bitstream = urlopen(bootloader_update_info[u"stage_one_url"]).read()

            print("    Programming stage one...")
            userimage_addr = p.meta.userimage_addr_range()[0]
            if p.program_bitstream(userimage_addr, bitstream):
                p.boot()
                print("    ...Success!")
            else:
                print("    ...Failure, try again!")
                return False
        else:
            print("    Update cancelled by user.")
            return False

    sys.stdout.write("    Waiting for stage one to reconnect...")
    sys.stdout.flush()

    new_port = None

    for x in range(20):
        time.sleep(1)
        sys.stdout.write(".")
        sys.stdout.flush()
        new_port = get_port_by_uuid("1d50:6130", uuid)
        if new_port is not None:
            print("connected!")
            break
    
    with new_port:
        p = TinyProg(new_port)

        ####################
        # STAGE TWO 
        ####################
        print("    Fetching stage two...")
        bitstream = urlopen(bootloader_update_info[u"stage_two_url"]).read()

        print("    Programming stage two...")
        if p.program_bitstream(0, bitstream):
            p.boot()
            print("    ...Success!\n")
            return True
        else:
            print("    ...Failure, try again!")
            return False


def print_board(port, m):
    if isinstance(m, dict) and u"boardmeta" in m:
        print("")
        print("        %s: %s %s" % (port, m[u"boardmeta"][u"name"], m[u"boardmeta"][u"hver"]))
        print("            UUID: %s" % m[u"boardmeta"][u"uuid"])
        print("            FPGA: %s" % m[u"boardmeta"][u"fpga"])
    else:
        print("")
        print("        %s: No metadata" % (port))


def main():
    def parse_int(str_value):
        str_value = str_value.strip().lower()
        if str_value.startswith("0x"):
            return int(str_value[2:], 16)
        else:
            return int(str_value)

    parser = argparse.ArgumentParser()

    parser.add_argument("-l", "--list", action="store_true",
                        help="list connected and active FPGA boards")
    parser.add_argument("-p", "--program", type=str,
                        help="program FPGA board with the given user bitstream")
    parser.add_argument("-u", "--program-userdata", type=str,
                        help="program FPGA board with the given user data")
    parser.add_argument("--program-image", type=str,
                        help="program FPGA board with a combined user bitstream and data")
    parser.add_argument("-b", "--boot", action="store_true",
                        help="command the FPGA board to exit the "
                             "bootloader and load the user configuration")
    parser.add_argument("-c", "--com", type=str, help="serial port name")
    parser.add_argument("-i", "--id", type=str, help="FPGA board ID")
    parser.add_argument("-d", "--device", type=str, default="1d50:6130",
                        help="device id (vendor:product); default is (1d50:6130)")
    parser.add_argument("-a", "--addr", type=str,
                        help="force the address to write the bitstream to")
    parser.add_argument("-m", "--meta", action="store_true",
                        help="dump out the metadata for all connected boards in JSON")
    parser.add_argument("--update-bootloader", action="store_true",
                        help="check for new bootloader and update eligible connected boards")
    parser.add_argument("--libusb", action="store_true",
                        help="try using libusb to connect to boards without a serial driver attached")
    parser.add_argument("--pyserial", action="store_true",
                        help="use pyserial to connect to boards")

    args = parser.parse_args()

    device = args.device.lower().replace(':', '')
    if len(device) != 8 or not all(c in '0123456789abcdef' for c in device):
        print("    Invalid device id, use format vendor:product")
        sys.exit(1)
    device = '{}:{}'.format(device[:4], device[4:])

    tinyprog.use_libusb = args.libusb
    tinyprog.use_pyserial = args.pyserial

    active_boards = get_ports(device) + get_ports("1209:2100")

    if args.meta:
        meta = []
        for port in active_boards:
            with port:
                p = TinyProg(port)
                m = p.meta.root
                m["port"] = str(port)
                meta.append(m)
        print(json.dumps(meta, indent=2))
        sys.exit(0)

    print("")
    print("    TinyProg CLI")
    print("    ------------")
    
    print("    Using device id {}".format(device))


    # find port to use
    active_port = None
    if args.com is not None:
        active_port = tinyprog.SerialPort(args.com)

    elif args.id is not None:
        active_port = get_port_by_uuid(device, args.id)

    elif not active_boards:
        print("    No port was specified and no active bootloaders found.")
        print("    Activate bootloader by pressing the reset button.")
        sys.exit(1)

    elif len(active_boards) == 1:
        print("    Only one board with active bootloader, using it.")
        active_port = active_boards[0]

    else:
        print("    Please choose a board with the -c or -i option.  Using first board in list.")
        active_port = active_boards[0]


    # list boards
    if args.list or active_port is None:
        print("    Boards with active bootloaders:")
        for port in active_boards:
            with port:
                p = TinyProg(port)
                m = p.meta.root
                print_board(port, m)

        if len(active_boards) == 0:
            print("        No active bootloaders found.  Check USB connections")
            print("        and press reset button to activate bootloader.")

    if args.update_bootloader:
        boards_needing_update = (
            check_for_wrong_tinyfpga_bx_vidpid() +
            check_for_new_bootloader()
        )

        if len(boards_needing_update) == 0:
            print("    All connected and active boards are up to date!")
        else:
            for port in boards_needing_update:
                perform_bootloader_update(port)

    # program the flash memory
    if (args.program is not None) or (args.program_userdata is not None) or (args.program_image is not None):
        boot_fpga = False
        def progress(info):
            if isinstance(info, str):
                print("    " + str(info))

        with active_port:
            fpga = TinyProg(active_port, progress)

            if args.program is not None:
                print("    Programming " + str(active_port) + " with " + str(args.program))

                bitstream = fpga.slurp(args.program)
                
                if args.addr is not None:
                    addr = parse_int(args.addr)
                else:
                    addr = fpga.meta.userimage_addr_range()[0]

                if addr < 0:
                    print("    Negative write addr: {}".format(addr))
                    sys.exit(1)
                if not fpga.is_bootloader_active():
                    print("    Bootloader not active")
                    sys.exit(1)

                if check_if_overwrite_bootloader(addr, len(bitstream), fpga.meta.userimage_addr_range()):
                    boot_fpga = True
                    print("    Programming at addr {:06x}".format(addr))
                    if not fpga.program_bitstream(addr, bitstream):
                        sys.exit(1)
    

            # program user flash area
            if args.program_userdata is not None:
                print("    Programming " + str(active_port) + " with " + str(args.program_userdata))

                bitstream = fpga.slurp(args.program_userdata)

                if args.addr is not None:
                    addr = parse_int(args.addr)
                else:
                    addr = fpga.meta.userdata_addr_range()[0]

                if addr < 0:
                    print("    Negative write addr: {}".format(addr))
                    sys.exit(1)
                if not fpga.is_bootloader_active():
                    print("    Bootloader not active")
                    sys.exit(1)

                if check_if_overwrite_bootloader(addr, len(bitstream), fpga.meta.userdata_addr_range()):
                    boot_fpga = True
                    print("    Programming at addr {:06x}".format(addr))
                    if not fpga.program_bitstream(addr, bitstream):
                        sys.exit(1)
    

            # program user image and data area
            if args.program_image is not None:
                print("    Programming " + str(active_port) + " with " + str(args.program_image))

                bitstream = fpga.slurp(args.program_image)

                if args.addr is not None:
                    addr = parse_int(args.addr)
                else:
                    addr = fpga.meta.userimage_addr_range()[0]

                if addr < 0:
                    print("    Negative write addr: {}".format(addr))
                    sys.exit(1)
                if not fpga.is_bootloader_active():
                    print("    Bootloader not active")
                    sys.exit(1)

                if check_if_overwrite_bootloader(addr, len(bitstream), (fpga.meta.userimage_addr_range()[0], fpga.meta.userdata_addr_range()[1])):
                    boot_fpga = True
                    print("    Programming at addr {:06x}".format(addr))
                    if not fpga.program_bitstream(addr, bitstream):
                        sys.exit(1)

            if boot_fpga:
                fpga.boot()
                print("")
                sys.exit(0)

    # boot the FPGA
    if args.boot:
        print("    Booting " + str(active_port))
        with active_port:
            fpga = TinyProg(active_port)
            fpga.boot()

    print("")


if __name__ == '__main__':
    main()



import struct
import json
import jsonmerge
import re
from intelhex import IntelHex
from tqdm import tqdm
from functools import reduce
import six
from serial.serialutil import SerialTimeoutException
from serial.tools.list_ports import comports
import serial
import platform

use_libusb = False
use_pyserial = False

def pretty_hex(data):
    output = ""
    for i in range(0, len(data), 16):
        output += " ".join(["%02x" % ord(x) for x in data[i:i+16]]) + "\n"
    return output

def to_int(value):
    try:
        return ord(value)
    except:
        return int(value)

def get_ports(device_id):
    """
    Return ports for all devices with the given device_id.

    :param device_id: USB VID and PID.
    :return: List of port objects.
    """

    ports = []

    if platform.system() == "Darwin" or use_libusb:
        import usb
        vid, pid = [int(x, 16) for x in device_id.split(":")]

        ports += [
            UsbPort(d)
            for d in usb.core.find(idVendor=vid, idProduct=pid, find_all=True)
            if not d.is_kernel_driver_active(1)
        ]

    # MacOS is not playing nicely with the serial drivers for the bootloader
    if platform.system() != "Darwin" or use_pyserial:
        # get serial ports first
        ports += [SerialPort(p[0]) for p in comports() if device_id in p[2].lower()]

    return ports


class SerialPort(object):
    def __init__(self, port_name):
        self.port_name = port_name
        self.ser = None

    def __str__(self):
        return self.port_name

    def __enter__(self):
        self.ser = serial.Serial(self.port_name, timeout=1.0, writeTimeout=1.0).__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.ser.__exit__(exc_type, exc_val, exc_tb)

    def write(self, data):
        self.ser.write(data)

    def flush(self):
        self.ser.flush()

    def read(self, length):
        return self.ser.read(length)


class UsbPort(object):
    def __init__(self, device):
        self.device = device
        usb_interface = device.configurations()[0].interfaces()[1]
        self.OUT = usb_interface.endpoints()[0]
        self.IN = usb_interface.endpoints()[1]

    def __str__(self):
        return "USB %d.%d" % (self.device.bus, self.device.port_number)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def write(self, data):
        self.OUT.write(data)

    def flush(self):
        # i don't think there's a comparable function on pyusb endpoints
        pass

    def read(self, length):
        if length > 0:
            data = self.IN.read(length)
            return bytearray(data)
        else:
            return ""


bit_reverse_table = bytearray([
    0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0, 0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0,
    0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8, 0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8,
    0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4, 0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4,
    0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC, 0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC,
    0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2, 0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2,
    0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA, 0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA,
    0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6, 0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6,
    0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE, 0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE,
    0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1, 0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1,
    0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9, 0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9,
    0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5, 0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5,
    0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED, 0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD,
    0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3, 0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3,
    0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB, 0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB,
    0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7, 0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7,
    0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF, 0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF
])


def _mirror_byte(b):
    return bit_reverse_table[to_int(b)]


def _mirror_each_byte(data):
    mirrored_data = bytearray(len(data))
    for i in range(0, len(data)):
        mirrored_data[i] = _mirror_byte(data[i])
    return mirrored_data


class TinyMeta(object):
    def __init__(self, prog):
        self.prog = prog
        prog.wake()
        self.root = self._read_metadata()

    def _parse_json(self, data):
        try:
            return json.loads(bytes(data).decode("utf-8"))
        except:
            return None

    def _resolve_pointers(self, meta):
        if isinstance(meta, dict):
            return {k: self._resolve_pointers(v) for k, v in meta.items()}

        if isinstance(meta, list):
            return [self._resolve_pointers(v) for v in meta]

        if isinstance(meta, (str, bytes)):
            m = re.search(r"^\s*@\s*0x(?P<addr>[A-Fa-f0-9]+)\s*\+\s*(?P<len>\d+)\s*$", meta)
            if m:
                data = self.prog.read(int(m.group("addr"), 16), int(m.group("len")))
                return json.loads(bytes(data).decode("utf-8"))
            else:
                return meta

        return meta

    def _read_metadata(self):
        import math
        meta_roots = (
            [self._parse_json(self.prog.read_security_register_page(p).replace(b"\x00", b"").replace(b"\xff", b"")) for p in [1, 2, 3]] +
            [self._parse_json(self.prog.read(int(math.pow(2, p) - (4 * 1024)), (4 * 1024)).replace(b"\x00", b"").replace(b"\xff", b"")) for p in [17, 18, 19, 20, 21, 22, 23, 24]]
        )
        meta_roots = [root for root in meta_roots if root is not None]
        if len(meta_roots) > 0:
            meta = reduce(jsonmerge.merge, meta_roots)
            return self._resolve_pointers(meta)
        return None

    def bootloader_addr_range(self):
        return self._get_addr_range(u"bootloader")

    def userimage_addr_range(self):
        return self._get_addr_range(u"userimage")

    def userdata_addr_range(self):
        return self._get_addr_range(u"userdata")

    def _get_addr_range(self, name):
        addr_str = self.root[u"bootmeta"][u"addrmap"][name]
        m = re.search(r"^\s*0x(?P<start>[A-Fa-f0-9]+)\s*-\s*0x(?P<end>[A-Fa-f0-9]+)\s*$", addr_str)
        if m:
            return int(m.group("start"), 16), int(m.group("end"), 16)
        else:
            return None

    def uuid(self):
        return str(self.root[u"boardmeta"][u"uuid"])



class TinyProg(object):
    def __init__(self, ser, progress=None):
        self.ser = ser

        if progress is None:
            self.progress = lambda x: x
        else:
            self.progress = progress

        self.wake()
        flash_id = self.read_id()
        flash_id = [to_int(b) for b in flash_id]
        # temporary hack, should have better database as well as SFPD reading
        if flash_id[0:2] == [0x9D, 0x60]:
            # ISSI
            self.security_page_bit_offset = 4
            self.security_page_write_cmd = 0x62
            self.security_page_read_cmd = 0x68
            self.security_page_erase_cmd = 0x64
    
        else:
            # Adesto
            self.security_page_bit_offset = 0
            self.security_page_write_cmd = 0x42
            self.security_page_read_cmd = 0x48
            self.security_page_erase_cmd = 0x44

        self.meta = TinyMeta(self)

    def is_bootloader_active(self):
        self.wake()
        devid = self.read_id()
        if devid not in [b'\xff\xff\xff']:
            return True
        return False

    def cmd(self, opcode, addr=None, data=b'', read_len=0):
        #print("type of data is %s" % type(data))
        #assert isinstance(data, bytes)
        addr = b'' if addr is None else struct.pack('>I', addr)[1:]
        write_string = bytearray([opcode]) + addr + data
        cmd_write_string = b'\x01' + struct.pack('<HH', len(write_string), read_len) + write_string
        self.ser.write(bytearray(cmd_write_string))
        self.ser.flush()
        return self.ser.read(read_len)

    def sleep(self):
        self.cmd(0xb9)

    def wake(self):
        self.cmd(0xab)

    def read_id(self):
        return self.cmd(0x9f, read_len=3)

    def read_sts(self):
        return self.cmd(0x05, read_len=1)

    def erase_security_register_page(self, page):
        self.write_enable()
        self.cmd(self.security_page_erase_cmd, page << (8 + self.security_page_bit_offset))
        self.wait_while_busy()

    def program_security_register_page(self, page, data):
        self.write_enable()
        self.cmd(self.security_page_write_cmd, page << (8 + self.security_page_bit_offset), data)
        self.wait_while_busy()

    def read_security_register_page(self, page):
        return self.cmd(self.security_page_read_cmd, addr=page << (8 + self.security_page_bit_offset), data=b'\x00', read_len=255)

    def read(self, addr, length, disable_progress=True, max_length = 255):
        data = b''
        with tqdm(desc="    Reading", unit="B", unit_scale=True, total=length, disable=disable_progress) as pbar:
            while length > 0:
                read_length = min(max_length, length)
                data += self.cmd(0x0b, addr, b'\x00', read_len=read_length)
                self.progress(read_length)
                addr += read_length
                length -= read_length
                pbar.update(read_length)
            return data

    def write_enable(self):
        self.cmd(0x06)

    def write_disable(self):
        self.cmd(0x04)

    def wait_while_busy(self):
        while to_int(self.read_sts()) & 1:
            pass

    def _erase(self, addr, length):
        opcode = {
             4 * 1024: 0x20,
            32 * 1024: 0x52,
            64 * 1024: 0xd8,
        }[length]
        self.write_enable()
        self.cmd(opcode, addr)
        self.wait_while_busy()

    def erase(self, addr, length, disable_progress=True):
        possible_lengths = (1, 4 * 1024, 32 * 1024, 64 * 1024)

        with tqdm(desc="    Erasing", unit="B", unit_scale=True, total=length, disable=disable_progress) as pbar:
            while length > 0:
                erase_length = max(p for p in possible_lengths
                                   if p <= length and addr % p == 0)

                if erase_length == 1:
                    # there are no opcode to erase that much
                    # either we want to erase up to multiple of 0x1000
                    # or we want to erase up to length

                    # start_addr                            end_addr
                    # v                                     v
                    # +------------------+------------------+----------------+
                    # |       keep       |      erase       |      keep      |
                    # +------------------+------------------+----------------+
                    #  <- start_length -> <- erase_length -> <- end_length ->

                    start_addr = addr & 0xfff000
                    start_length = addr & 0xfff
                    erase_length = min(0x1000 - start_length, length)
                    end_addr = start_addr + start_length + erase_length
                    end_length = start_addr + 0x1000 - end_addr

                    # read data we need to restore later
                    if start_length:
                        start_read_data = self.read(start_addr, start_length)
                    if end_length:
                        end_read_data = self.read(end_addr, end_length)

                    # erase the block
                    self._erase(start_addr, 0x1000)

                    # restore data
                    if start_length:
                        self.write(start_addr, start_read_data)
                    if end_length:
                        self.write(end_addr, end_read_data)

                else:
                    # there is an opcode to erase that much data
                    self.progress(erase_length)
                    self._erase(addr, erase_length)

                # update
                length -= erase_length
                addr += erase_length
                pbar.update(erase_length)

    # don't use this directly, use the public "write" function instead
    def _write(self, addr, data):
        self.write_enable()
        self.cmd(0x02, addr, data)
        self.wait_while_busy()
        self.progress(len(data))

    def write(self, addr, data, disable_progress=True, max_length = 256):
        offset = 0
        with tqdm(desc="    Writing", unit="B", unit_scale=True, total=len(data), disable=disable_progress) as pbar:
            while offset < len(data):
                dist_to_256_byte_boundary = 256 - (addr & 0xff)
                write_length = min(max_length, len(data) - offset, dist_to_256_byte_boundary)
                write_data = data[offset : offset+write_length]
            
                self._write(addr, write_data)
                offset += write_length
                addr += write_length
                pbar.update(write_length)

    def program_fast(self, addr, data):
        self.erase(addr, len(data), disable_progress=False)
        self.write(addr, data, disable_progress=False)
        read_back = self.read(addr, len(data), disable_progress=False)

        if read_back == data:
            self.progress("Success!")
            return True
        else:
            read_back_file = open("readback.bin", "wb")
            read_back_file.write(read_back)
            read_back_file.close()
            self.progress("FAILED!")
            return False

    def program_sectors(self, addr, data):
        sector_size = 4 * 1024

        with tqdm(desc="    Programming and Verifying", unit="B", unit_scale=True, total=len(data)) as pbar:
            for offset in range(0, len(data), sector_size):
                current_addr = addr + offset
                current_write_data = data[offset:offset + sector_size]
                self.erase(current_addr, sector_size, disable_progress=True)

                minor_sector_size = 256
                for minor_offset in range(0, 4 * 1024, minor_sector_size):
                    minor_write_data = current_write_data[minor_offset:minor_offset+minor_sector_size]
                    self.write(current_addr + minor_offset, minor_write_data, disable_progress=True, max_length=256)
                    minor_read_data = self.read(current_addr + minor_offset, len(minor_write_data), disable_progress=True, max_length=255)
                    
                    pbar.update(len(minor_write_data))

                    if minor_read_data != minor_write_data:
                        print("")
                        print("Offset: %d" % (current_addr + minor_offset))
                        print("Readback Data:")
                        print(pretty_hex(minor_read_data))
                        print("Write Data:")
                        print(pretty_hex(minor_write_data))
                        self.progress("FAILED!")
                        return False


        self.progress("Success!")
        return True


    def boot(self):
        try:
            self.ser.write(b"\x00")
            self.ser.flush()
        except:
            # we might get a writeTimeoutError and that's OK.  Sometimes the
            # bootloader will reboot before it finishes sending out the USB ACK
            # for the boot command data packet.
            pass

    def slurp(self, filename):
        if filename.endswith('.bit') or filename.endswith('.bin'):
            with open(filename, 'rb') as f:
                return f.read()

        elif filename.endswith('.hex'):
            with open(filename, 'rb') as f:
                return bytes("".join(chr(int(i, 16)) for i in f.read().split()))

        elif filename.endswith('.mcs'):
            ih = IntelHex()
            ih.fromfile(filename, format='hex')
            bitstream = ih.tobinstr(start=0, end=ih.maxaddr())
            bitstream = _mirror_each_byte(bitstream)
            return bitstream

        else:
            raise ValueError('Unknown bitstream extension')

    def program_bitstream(self, addr, bitstream):
        self.progress("Waking up SPI flash")
        self.wake()
        self.progress(str(len(bitstream)) + " bytes to program")
        return self.program_sectors(addr, bitstream)

    
    def program_bitstream_fast(self, addr, bitstream):
        self.progress("Waking up SPI flash")
        self.wake()
        self.progress(str(len(bitstream)) + " bytes to program")
        return self.program_fast(addr, bitstream)
